{% extends "base.html" %}

{% block announce %}
  <strong>⚠️ mkdocstrings is in maintenance mode!
    <a href="https://pawamoy.github.io/posts/sunsetting-the-sponsorware-strategy/">blog post</a>
  </strong>
{% endblock %}
